package PScm;

use strict;
use warnings;
use PScm::Read;
use PScm::Env;
use PScm::Primitive;
use PScm::SpecialForm;
use PScm::Class;
use FileHandle;

require Exporter;

our @ISA    = qw(Exporter);
our @EXPORT = qw(ReadEvalPrint);

=head1 NAME

PScm - PScm interpreter written in Perl

=head1 SYNOPSIS

  use PScm;
  ReadEvalPrint($in_filehandle[, $out_filehandle]);

=head1 DESCRIPTION

Just messing about, A toy lisp interpreter.

=cut

sub ReadEvalPrint {
    my ($infh, $outfh) = @_;

    $outfh ||= new FileHandle(">-");
    my $reader      = new PScm::Read($infh);
    my $initial_env = new PScm::Env(
        let          => new PScm::SpecialForm::Let(),
        '*'          => new PScm::Primitive::Multiply(),
        '-'          => new PScm::Primitive::Subtract(),
        '+'          => new PScm::Primitive::Add(),
        if           => new PScm::SpecialForm::If(),
        lambda       => new PScm::SpecialForm::Lambda(),
        list         => new PScm::Primitive::List(),
        car          => new PScm::Primitive::Car(),
        cdr          => new PScm::Primitive::Cdr(),
        cons         => new PScm::Primitive::Cons(),
        letrec       => new PScm::SpecialForm::LetRec(),
        'let*'       => new PScm::SpecialForm::LetStar(),
        eval         => new PScm::SpecialForm::Eval(),
        macro        => new PScm::SpecialForm::Macro(),
        quote        => new PScm::SpecialForm::Quote(),
        'set!'       => new PScm::SpecialForm::Set(),
        begin        => new PScm::SpecialForm::Begin(),
        define       => new PScm::SpecialForm::Define(),
        'make-class' => new PScm::SpecialForm::MakeClass(),
    );

    $initial_env->Define(
        PScm::Expr::Symbol->new("root"),
        PScm::Class::Root->new($initial_env)
    );

    while (defined(my $expr = $reader->Read)) {
        my $result = $expr->Eval($initial_env);
        $result->Print($outfh);
    }
}

sub Print {
    my ($expr, $outfh) = @_;
    print $outfh $expr->as_string, "\n";
}

sub as_string { ref($_[0]); }

sub new { bless {}, $_[0] }

1;
