package PScm::Primitive;

use strict;
use warnings;
use base qw(PScm::Expr);

sub Apply {
    my ($self, $form, $env) = @_;

    my @evaluated_args = $form->map_eval($env)->value();
    return $self->_apply(@evaluated_args);
}

sub _check_type {
    my ($self, $thing, $type) = @_;

    die "wrong type argument(", ref($thing), ") to ", ref($self),
      "\n"
      unless $thing->isa($type);
}

##################################
package PScm::Primitive::Multiply;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, @args) = @_;

    my $result = PScm::Expr::Number->new(1)->value();

    while (@args) {
        my $arg = shift @args;
        $self->_check_type($arg, 'PScm::Expr::Number');
        $result *= $arg->value;
    }

    return new PScm::Expr::Number($result);
}

##################################
package PScm::Primitive::Subtract;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, @args) = @_;

    unshift @args, PScm::Expr::Number->new(0) if @args < 2;

    my $arg = shift @args;
    $self->_check_type($arg, 'PScm::Expr::Number');

    my $result = $arg->value;

    while (@args) {
        $arg = shift @args;
        $self->_check_type($arg, 'PScm::Expr::Number');
        $result -= $arg->value;
    }

    return new PScm::Expr::Number($result);
}

##################################
package PScm::Primitive::Add;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, @args) = @_;

    my $result = 0;

    while (@args) {
        my $arg = shift @args;
        $self->_check_type($arg, 'PScm::Expr::Number');
        $result += $arg->value;
    }

    return new PScm::Expr::Number($result);
}

##############################
package PScm::Primitive::List;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, @args) = @_;

    return new PScm::Expr::List(@args);
}

#############################
package PScm::Primitive::Car;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, $arg) = @_;

    $self->_check_type($arg, 'PScm::Expr::List');
    return $arg->first;
}

#############################
package PScm::Primitive::Cdr;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, $arg) = @_;

    $self->_check_type($arg, 'PScm::Expr::List');
    return $arg->rest;
}

##############################
package PScm::Primitive::Cons;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, $car, $cdr) = @_;

    return PScm::Expr::List->Cons($car, $cdr);
}

1;
